import argparse
import os
import sys
from datetime import datetime, timedelta


class TaskCronConfig:

    def __init__(self, filepath: str, hour: int | None, minute: int | None):
        self._hour = hour
        self._minute = minute
        self._filepath = filepath

    @property
    def filepath(self):
        return self._filepath

    @property
    def hour(self):
        return self._hour

    @property
    def minute(self):
        return self._minute


class TaskSoonestExecutionTime:
    def __init__(self, task_cron_config: TaskCronConfig, current_time: datetime):
        self._task_cron_config = task_cron_config
        self._current_time = current_time
        self._soonest_execution_time = self._calc_next_exec_time()

    def _calc_next_exec_time(self) -> datetime:
        current_time = self._current_time
        now = datetime.now().replace(hour=current_time.hour, minute=current_time.minute)
        tomorrow_same_time = now + timedelta(days=1)
        task_hour = self._task_cron_config.hour
        task_minute = self._task_cron_config.minute
        if task_hour is None and task_minute is None:
            self._soonest_execution_time = now
        elif task_hour is None and task_minute is not None:
            if current_time.minute <= task_minute:
                # Still time left within that minute for executing the task
                return now.replace(minute=task_minute)
            else:
                # No time left, will be executed next hour
                return now.replace(hour=current_time.hour + 1, minute=task_minute)
        elif task_minute is None and task_hour is not None:
            if current_time.hour < task_hour:
                return now.replace(hour=task_hour, minute=0)
            elif current_time.hour == task_hour:
                return now
            else:
                return tomorrow_same_time.replace(hour=task_hour, minute=0)
        else:
            assert task_minute is not None and task_hour is not None
            if current_time.hour <= task_hour and current_time.minute <= task_minute:
                return now.replace(hour=task_hour, minute=task_minute)
            else:
                return tomorrow_same_time.replace(hour=task_hour, minute=task_minute)

    @ property
    def soonest_execution_time(self) -> datetime:
        return self._soonest_execution_time

    @ property
    def filepath(self) -> str:
        return self._task_cron_config.filepath


def parse_time() -> datetime:
    """
    Parse time argument
    :raises ValueError if argument passed isn't in HH:mm format
    """
    parser = argparse.ArgumentParser(
        description='Calculate the soonest time each cron commands will be executed at.')
    parser.add_argument('current_time', type=str, required=True,
                        help='Current Time in HH:mm format')
    args = parser.parse_args()
    return datetime.strptime(args.current_time, '%H:%M')


def read_stdin_and_validate(current_time: datetime) -> list[TaskCronConfig]:
    """
    :raises ValueError if data in stdin isn't a valid config file
    """
    if sys.stdin is None:
        raise ValueError('Please feed a minicron config file to stdin')
    task_cron_configs = list()
    for line in sys.stdin:
        # Remove '\n' character
        line = line.rstrip()
        # Get each column in separate variables:
        groups: list[str] = line.split(' ')
        if len(groups) != 3:
            raise ValueError(
                'The minicron config should have 3 whitespace separated columns per line')
        # Using None for '*' is not pretty nor typesafe but for the purpose of this exercise it will do the trick
        # The following can raise a ValueError if validation fails
        minute = None if groups[0] == '*' else datetime.strptime(
            groups[0], '%M').minute
        hour = None if groups[1] == '*' else datetime.strptime(
            groups[1], '%H').hour
        filepath = groups[2]
        if not os.path.isfile(filepath):
            raise ValueError("File path '{}' does not exist".format(filepath))
        task_cron_configs.append(TaskCronConfig(filepath=filepath,
                                                hour=hour,
                                                minute=minute))
    return task_cron_configs


def calc_and_print_soonest_time():
    try:
        current_time: datetime = parse_time()
    except ValueError as e:
        print('Cannot parse current time: {}'.format(e))
        return
    else:
        try:
            task_cron_configs = read_stdin_and_validate()
        except ValueError as e:
            print('Cannot parse cron config: {}'.format(e))
        else:
            calculate_next_executions(task_cron_configs, current_time)


if __name__ == "__main__":
    calc_and_print_soonest_time()
